@@ -10,9 +10,6 @@ from itertools import zip_longest
 import importlib.resources as pkg_resources
 import xml.etree.ElementTree as ET
 
-from latex2mathml.converter import convert  # type: ignore
-import latex2mathml.commands  # type: ignore
-
 import ziafont as zf
 from ziafont.glyph import fmt
 from .mathfont import MathFont
@@ -20,6 +17,7 @@ from .nodes import Mnode
 from .styles import parse_style
 from .escapes import unescape
 from .config import config
+from .tex import tex2mml
 
 
 Halign = Literal['left', 'center', 'right']
@@ -37,37 +35,6 @@ def denamespace(element: ET.Element) -> ET.Element:
     return element
 
 
-def declareoperator(name: str) -> None:
-    r''' Declare a new operator name, similar to Latex ``\DeclareMathOperator`` command.
-
-        Args:
-            name: Name of operator, should start with a ``\``.
-                Example: ``declareoperator(r'\myfunc')``
-    '''
-    latex2mathml.commands.FUNCTIONS = latex2mathml.commands.FUNCTIONS + (name,)
-
-
-def tex2mml(tex: str, inline: bool = False) -> str:
-    ''' Convert Latex to MathML. Do some hacky preprocessing to work around
-        some issues with generated MathML that ziamath doesn't support yet.
-    '''
-    tex = re.sub(r'\\binom{(.+?)}{(.+?)}', r'\\left( \1 \\atop \2 \\right)', tex)
-    # latex2mathml bug requires space after mathrm
-    tex = re.sub(r'\\mathrm{(.+?)}', r'\\mathrm {\1}', tex)
-    tex = tex.replace('||', '‖')
-    if config.decimal_separator == ',':
-        # Replace , with {,} to remove right space
-        # (must be surrounded by digits)
-        tex = re.sub(r'([0-9]),([0-9])', r'\1{,}\2', tex)
-
-    mml = convert(tex, display='inline' if inline else 'block')
-
-    # Replace some operators with "stretchy" variants
-    mml = re.sub(r'<mo>&#x0005E;', r'<mo>&#x00302;', mml)  # widehat
-    mml = re.sub(r'<mo>&#x0007E;', r'<mo>&#x00303;', mml)  # widetilde
-    return mml
-
-
 def apply_mstyle(element: ET.Element) -> ET.Element:
     ''' Take attributes defined in <mstyle> elements and add them
         to all the child elements, removing the original <mstyle>
@@ -95,7 +62,7 @@ class Math:
             font: Filename of font file. Must contain MATH typesetting table.
     '''
     def __init__(self, mathml: Union[str, ET.Element],
-                 size: float = None, font: str = None):
+                 size: Optional[float] = None, font: Optional[str] = None):
         self.size = size if size else config.math.fontsize
         font = font if font else config.math.mathfont
 
@@ -121,8 +88,8 @@ class Math:
         self.node = Mnode.fromelement(mathml, parent=self)  # type: ignore
 
     @classmethod
-    def fromlatex(cls, latex: str, size: float = None, mathstyle: str = None,
-                  font: str = None, color: str = None, inline: bool = False):
+    def fromlatex(cls, latex: str, size: Optional[float] = None, mathstyle: Optional[str] = None,
+                  font: Optional[str] = None, color: Optional[str] = None, inline: bool = False):
         ''' Create Math Renderer from a single LaTeX expression. Requires
             latex2mathml Python package.
 
@@ -146,9 +113,9 @@ class Math:
         return cls(mathml, size, font)
 
     @classmethod
-    def fromlatextext(cls, latex: str, size: float = 24, mathstyle: str = None,
-                      textstyle: str = None, font: str = None,
-                      color: str = None):
+    def fromlatextext(cls, latex: str, size: float = 24, mathstyle: Optional[str] = None,
+                      textstyle: Optional[str] = None, font: Optional[str] = None,
+                      color: Optional[str] = None):
         ''' Create Math Renderer from a sentence containing zero or more LaTeX
             expressions delimited by $..$, resulting in single MathML element.
             Requires latex2mathml Python package.
@@ -162,7 +129,7 @@ class Math:
                 color: Color parameter, equivalent to "mathcolor" attribute
         '''
         warnings.warn(r'fromlatextext is deprecated. Use ziamath.Text or \text{} command.', DeprecationWarning, stacklevel=2)
-        
+
         # Extract each $..$, convert to MathML, but the raw text in <mtext>, and join
         # into a single <math>
         parts = re.split(r'(\$+.*?\$+)', latex)
@@ -250,7 +217,7 @@ class Math:
 
     @classmethod
     def mathml2svg(cls, mathml: Union[str, ET.Element],
-                   size: float = None, font: str = None):
+                   size: Optional[float] = None, font: Optional[str] = None):
         ''' Shortcut to just return SVG string directly '''
         return cls(mathml, size=size, font=font).svg()
 
@@ -275,8 +242,8 @@ class Latex(Math):
             color: Color parameter, equivalent to "mathcolor" attribute
             inline: Use inline math mode (default is block mode)
         '''
-    def __init__(self, latex: str, size: float = None, mathstyle: str = None,
-                 font: str = None, color: str = None, inline: bool = False):
+    def __init__(self, latex: str, size: Optional[float] = None, mathstyle: Optional[str] = None,
+                 font: Optional[str] = None, color: Optional[str] = None, inline: bool = False):
         self.latex = latex
 
         mathml: Union[str, ET.Element]
@@ -312,9 +279,9 @@ class Text:
                 https://matplotlib.org/stable/gallery/text_labels_and_annotations/demo_text_rotation_mode.html
 
     '''
-    def __init__(self, s, textfont: str = None, mathfont: str = None,
-                 mathstyle: str = None, size: float = None, linespacing: float = None,
-                 color: str = None,
+    def __init__(self, s, textfont: Optional[str] = None, mathfont: Optional[str] = None,
+                 mathstyle: Optional[str] = None, size: Optional[float] = None, linespacing: Optional[float] = None,
+                 color: Optional[str] = None,
                  halign: str = 'left', valign: str = 'base',
                  rotation: float = 0, rotation_mode: str = 'anchor'):
         self.str = s
@@ -337,10 +304,13 @@ class Text:
         if textfont is None:
             self.textfont = None
             self.textstyle = textstyle
+        elif textfont in loadedtextfonts:
+            self.textfont = loadedtextfonts[textfont]
         else:
             try:
                 self.textfont = zf.Font(textfont)
                 self.textstyle = textstyle
+                loadedtextfonts[str(textfont)] = self.textfont
             except FileNotFoundError:
                 self.textfont = None
                 self.textstyle = textfont
@@ -371,7 +341,7 @@ class Text:
             f.write(self.svg())
 
     def drawon(self, svg: ET.Element, x: float = 0, y: float = 0,
-               halign: str = None, valign: str = None) -> ET.Element:
+               halign: Optional[str] = None, valign: Optional[str] = None) -> ET.Element:
         ''' Draw text on existing SVG element
 
             Args:
@@ -385,7 +355,7 @@ class Text:
         return svgelm
 
     def _drawon(self, svg: ET.Element, x: float = 0, y: float = 0,
-                halign: str = None, valign: str = None) -> Tuple[ET.Element, Tuple[float, float, float, float]]:
+                halign: Optional[str] = None, valign: Optional[str] = None) -> Tuple[ET.Element, Tuple[float, float, float, float]]:
         ''' Draw text on existing SVG element
 
             Args:
@@ -429,20 +399,22 @@ class Text:
                     svgparts.append(math)
                     partsizes.append(math.getsize())
                 else:  # Text
+                    part = part.replace('<', '&lt;').replace('>', '&gt;')
                     if self.textfont:
                         # A specific font file is defined, use ziafont and ignore textstyle
                         txt = zf.Text(part, font=self.textfont, size=self.size, color=self.textcolor)
                         partsizes.append(txt.getsize())
+                        svgparts.append(txt)
                     else:
                         # use math font with textstyle
-                        txt = Math.fromlatex(fr'\text{{{part}}}',
+                        txt = Math.fromlatex(f'\\text{{{part}}}',
                                              font=self.mathfont,
                                              mathstyle=self.textstyle,
                                              size=self.size,
                                              color=self.textcolor)
-                        partsizes.append((txt.node.bbox.xmax - txt.node.bbox.xmin,  # type: ignore
-                                          txt.node.size))  # type: ignore
+                        partsizes.append(txt.getsize())
                     svgparts.append(txt)
+
             if len(svgparts) > 0:
                 svglines.append(svgparts)
                 linesizes.append(partsizes)
@@ -452,11 +424,11 @@ class Text:
         linewidths = [sum(p[0] for p in line) for line in linesizes]
 
         if valign == 'bottom':
-            ystart = y + sum(lineofsts) - sum(lineheights[1:])
+            ystart = y - (len(lines)-1)*self.size*self.linespacing + lineofsts[-1]
         elif valign == 'top':
             ystart = y + lineheights[0] + lineofsts[0]
         elif valign == 'center':
-            ystart = y + lineheights[0] + lineofsts[0] - sum(lineheights)/2
+            ystart = y + (lineheights[0] + lineofsts[0] - (len(lines)-1)*self.size*self.linespacing + lineofsts[-1])/2
         else:  # 'base'
             ystart = y
 
@@ -472,22 +444,13 @@ class Text:
             xmin = min(xmin, xloc)
             xmax = max(xmax, xloc+linewidths[i])
 
-            # Include extra height of tall math expressions
-            drop = 0
-            if i > 0:
-                try:
-                    drop = min(p.node.bbox.ymin for p in line if isinstance(p, Math))
-                except ValueError:
-                    drop = 0
-
-            yloc -= drop
-            ymin = min(ymin, yloc-lineheights[i])
-            ymax = max(ymax, yloc-lineofsts[i])
             for part, size in zip(line, linesizes[i]):
                 part.drawon(svgelm, xloc, yloc)
                 xloc += size[0]
-            yloc += lineheights[i] * self.linespacing
+            yloc += self.size * self.linespacing
 
+        ymin = y - lineheights[0] - lineofsts[0]
+        ymax = yloc - self.size*self.linespacing - lineofsts[-1]
         if self.rotation:
             costh = cos(radians(self.rotation))
             sinth = sin(radians(self.rotation))
@@ -534,20 +497,25 @@ class Text:
 
         return svgelm, (xmin, xmax, ymin, ymax)
 
-    def getsize(self):
+    def getsize(self) -> tuple[float, float]:
         ''' Get pixel width and height of Text. '''
         svg = ET.Element('svg')
         _, (xmin, xmax, ymin, ymax) = self._drawon(svg)
         return (xmax-xmin, ymax-ymin)
 
-    def bbox(self):
+    def bbox(self) -> tuple[float, float, float, float]:
         ''' Get bounding box (xmin, xmax, ymin, ymax) of Text. '''
         svg = ET.Element('svg')
         _, (xmin, xmax, ymin, ymax) = self._drawon(svg)
         return (xmin, xmax, ymin, ymax)
 
+    def getyofst(self) -> float:
+        ''' Y offset from baseline to bottom of bounding box '''
+        return -self.bbox()[3]
+
 
 # Cache the loaded fonts to prevent reloading all the time
 with pkg_resources.path('ziamath.fonts', 'STIXTwoMath-Regular.ttf') as p:
     fontname = p
 loadedfonts: Dict[str, MathFont] = {'default': MathFont(fontname)}
+loadedtextfonts: Dict[str, zf.Font] = {}
\ No newline at end of file
